@@ -5,7 +5,9 @@ import shutil
 from spur.tempdir import create_temporary_dir
 from spur.files import FileOperations
 import spur.results
-    
+from .io import IoHandler
+
+
 class LocalShell(object):
     def upload_dir(self, source, dest, ignore=None):
         shutil.copytree(source, dest, ignore=shutil.ignore_patterns(*ignore))
@@ -21,16 +23,19 @@ class LocalShell(object):
         open(remote_path, "w").write(contents)
 
     def spawn(self, *args, **kwargs):
+        stdout = kwargs.pop("stdout", None)
+        stderr = kwargs.pop("stderr", None)
+        allow_error = kwargs.pop("allow_error", False)
         process = subprocess.Popen(**self._subprocess_args(*args, **kwargs))
-        return LocalProcess(process)
+        return LocalProcess(
+            process,
+            allow_error=allow_error,
+            stdout=stdout,
+            stderr=stderr
+        )
         
     def run(self, *args, **kwargs):
-        allow_error = kwargs.pop("allow_error", False)
-        process = self.spawn(*args, **kwargs)
-        return spur.results.result(
-            process,
-            allow_error=allow_error
-        )
+        return self.spawn(*args, **kwargs).wait_for_result()
         
     def temporary_dir(self):
         return create_temporary_dir()
@@ -56,9 +61,15 @@ class LocalShell(object):
         return kwargs
 
 class LocalProcess(object):
-    def __init__(self, subprocess):
+    def __init__(self, subprocess, allow_error, stdout, stderr):
         self._subprocess = subprocess
+        self._allow_error = allow_error
         self._result = None
+            
+        self._io = IoHandler([
+            (subprocess.stdout, stdout),
+            (subprocess.stderr, stderr),
+        ], lambda: self._subprocess.communicate())
         
     def is_running(self):
         return self._subprocess.poll() is None
@@ -71,12 +82,15 @@ class LocalProcess(object):
             self._result = self._generate_result()
             
         return self._result
-        
+    
     def _generate_result(self):
-        stdout, stderr = self._subprocess.communicate()
+        output, stderr_output = self._io.wait()
         return_code = self._subprocess.poll()
-        return spur.results.ExecutionResult(
+        
+        return spur.results.result(
             return_code,
-            stdout,
-            stderr
+            self._allow_error,
+            output,
+            stderr_output
         )
+
